from .enums import Operator


class Condition:
    def __init__(self, field: str, value: any, operator: Operator = Operator.EQ):
        self.field: str = field
        self.value: any = value
        self.operator: Operator = operator

    def parse(self):
        return f'{self.field} {self.operator.value} ?', self.value


class ConditionTree:
    def __init__(self, logic='and'):
        self.conditions = []
        self.logic = logic

    def count(self):
        return len(self.conditions)

    def add_condition(self, condition: Condition):
        self.conditions.append(condition)
        return self

    def add_tree(self, condition_tree):
        self.conditions.append(condition_tree)
        return self

    def parse(self):
        if len(self.conditions) == 0:
            return None
        args = []
        exps = []
        for condition in self.conditions:
            if isinstance(condition, ConditionTree):
                exp, arg = condition.parse()
                exps.append(f'({exp})')
                args.extend(arg)
            else:
                exp, arg = condition.parse()
                exps.append(exp)
                args.append(arg)
        return f' {self.logic} '.join(exps), tuple(args)



from dataclasses import make_dataclass
from typing import get_type_hints


def entity(table: str):
    def _entity(cls):
        annotations = get_type_hints(cls)
        new_fields = []
        for name, typ in annotations.items():
            if name.startswith('__'):
                continue
            new_fields.append((name, typ, None))
        new_cls = make_dataclass(cls.__name__, new_fields, bases=cls.__bases__)
        new_cls.__table_name__ = table
        for new_field in new_fields:
            setattr(new_cls, new_field[0], new_field[0])
        return new_cls

    return _entity



from dataclasses import fields
from typing import List, Any, TypeVar, Type, Generic

from ._dorm import dorm
from ._models import models
from ._mysql_executor import select_one, select_many
from ._where import Where, Or
from .protocols import IDataSource

T = TypeVar('T')


class RawQuery:
    def __init__(self, table: str, datasource: IDataSource, database: str):
        self._table = table
        self._datasource = datasource
        self._database = database

        if self._table is None or self._table == '':
            raise ValueError('table is required')
        if self._datasource is None or self._datasource == '':
            raise ValueError('datasource is required')
        if not isinstance(self._datasource, IDataSource):
            raise ValueError('datasource must be an instance of IDataSource')
        if self._database is None or self._database == '':
            raise ValueError('database is required')

        self._where = Where()
        self._select_fields = []
        self._ignore_fields = []
        self._order_by = None
        self._limit = None
        self._offset = None
        self._distinct = False

        self._model: callable = models.get(data_source=self._datasource, database=self._database, table=self._table)
        self._model_fields = [f.name for f in fields(self._model)]

    def select(self, *select_fields, distinct=False) -> 'RawQuery':
        model_fields = [f.name for f in fields(self._model)]
        for select_field in select_fields:
            if select_field not in model_fields:
                raise ValueError(f'invalid field [{select_field}]')

        self._select_fields = select_fields
        self._distinct = distinct
        return self

    def ignore(self, *ignore_fields) -> 'RawQuery':
        model_fields = [f.name for f in fields(self._model)]
        for ignore_field in ignore_fields:
            if ignore_field not in model_fields:
                raise ValueError(f'invalid field [{ignore_field}]')

        self._ignore_fields = ignore_fields
        return self

    def check_field(self, field: str):
        if field not in self._model_fields:
            raise ValueError(f'invalid field [{field}]')

    def eq(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.eq(field, value)
        return self

    def ne(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.ne(field, value)
        return self

    def gt(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.gt(field, value)
        return self

    def ge(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.ge(field, value)
        return self

    def lt(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.lt(field, value)
        return self

    def le(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.le(field, value)
        return self

    def in_(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.in_(field, value)
        return self

    def l_like(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.l_like(field, value)
        return self

    def r_like(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.r_like(field, value)
        return self

    def like(self, field: str, value: Any) -> 'RawQuery':
        self.check_field(field)
        self._where.like(field, value)
        return self

    def or_(self, or_: Or) -> 'RawQuery':
        self._where.or_(or_)
        return self

    def desc(self, *order_by) -> 'RawQuery':
        self._order_by = [f'{field} desc' for field in order_by]
        return self

    def asc(self, *order_by) -> 'RawQuery':
        self._order_by = [f'{field} asc' for field in order_by]
        return self

    def limit(self, limit: int) -> 'RawQuery':
        self._limit = limit
        return self

    def offset(self, offset: int) -> 'RawQuery':
        self._offset = offset
        return self

    def one(self) -> dict[str, Any] | None:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._model)]

        sql, args = self._build_select()
        conn = self._datasource.get_connection()
        return select_one(sql, args, conn)

    def list(self) -> list[dict[str, Any]]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._model)]

        sql, args = self._build_select()
        conn = self._datasource.get_connection()
        rows = select_many(sql, args, conn)
        if rows is None:
            return []
        return rows

    def page(self, page: int, page_size: int) -> tuple[List[dict[str, Any]], int]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._model)]

        self._limit = page_size
        self._offset = (page - 1) * page_size
        sql, args = self._build_select()
        conn = self._datasource.get_connection()
        count = self.count()
        if count == 0:
            return [], count
        rows = select_many(sql, args, conn)
        if rows is None:
            return [], count
        return rows, count

    def count(self) -> int:
        sql = f'SELECT COUNT(*) FROM {self._database}.{self._table}'
        args = ()
        tree = self._where.tree()
        if len(tree.conditions) > 0:
            exp, args = tree.parse()
            sql += ' WHERE ' + exp
        conn = self._datasource.get_connection()
        return select_one(sql, args, conn)['COUNT(*)']

    def _build_select(self) -> tuple[str, tuple[Any, ...]]:
        select_fields = [field for field in self._select_fields if field not in self._ignore_fields]

        sql = f'SELECT {"DISTINCT " if self._distinct and self._select_fields else ""}{",".join(select_fields)} FROM {self._database}.{self._table}'
        args = ()
        tree = self._where.tree()
        if len(tree.conditions) > 0:
            exp, args = tree.parse()
            sql += ' WHERE ' + exp
        if self._order_by is not None:
            sql += f' ORDER BY {",".join(self._order_by)}'
        if self._limit is not None:
            sql += f' LIMIT {self._limit}'
        if self._offset is not None:
            sql += f' OFFSET {self._offset}'

        return sql, args


class Query(RawQuery, Generic[T]):
    def __init__(self, cls: Type[T], datasource: IDataSource, database: str):
        if not hasattr(cls, '__table_name__'):
            raise ValueError('invalid model class')
        super().__init__(cls.__table_name__, datasource, database)
        self._cls = cls

    def eq(self, field: str, value: Any) -> 'Query':
        super().eq(field, value)
        return self

    def ne(self, field: str, value: Any) -> 'Query':
        super().ne(field, value)
        return self

    def gt(self, field: str, value: Any) -> 'Query':
        super().gt(field, value)
        return self

    def ge(self, field: str, value: Any) -> 'Query':
        super().ge(field, value)
        return self

    def lt(self, field: str, value: Any) -> 'Query':
        super().lt(field, value)
        return self

    def le(self, field: str, value: Any) -> 'Query':
        super().le(field, value)
        return self

    def in_(self, field: str, value: Any) -> 'Query':
        super().in_(field, value)
        return self

    def l_like(self, field: str, value: Any) -> 'Query':
        super().l_like(field, value)
        return self

    def r_like(self, field: str, value: Any) -> 'Query':
        super().r_like(field, value)
        return self

    def like(self, field: str, value: Any) -> 'Query':
        super().like(field, value)
        return self

    def or_(self, or_: Or) -> 'Query':
        super().or_(or_)
        return self

    def desc(self, *order_by) -> 'Query':
        super().desc(*order_by)
        return self

    def asc(self, *order_by) -> 'Query':
        super().asc(*order_by)
        return self

    def limit(self, limit: int) -> 'Query':
        self._limit = limit
        return self

    def offset(self, offset: int) -> 'Query':
        self._offset = offset
        return self

    def one(self) -> T | None:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._cls)]

        row = super().one()
        if row is None:
            return None
        return self._cls(**row)

    def list(self) -> List[T]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._cls)]

        rows = super().list()
        if len(rows) == 0:
            return rows
        return [self._cls(**row) for row in rows]

    def page(self, page: int, page_size: int) -> tuple[List, int]:
        if len(self._select_fields) == 0:
            self._select_fields = [f.name for f in fields(self._cls)]

        rows, count = super().page(page, page_size)
        if len(rows) == 0:
            return rows, count
        return [self._cls(**row) for row in rows], count


def query(cls: Type[T], database: str | None = None, data_source: IDataSource | None = None):
    return Query(cls, data_source or dorm.default_datasource(), database or dorm.default_datasource().get_default_database())


def raw_query(table: str, database: str | None = None, data_source: IDataSource | None = None):
    return RawQuery(table, data_source or dorm.default_datasource(), database or dorm.default_datasource().get_default_database())



from dataclasses import field, make_dataclass

from loguru import logger

from pydorm.protocols import IDataSource


class Models:
    def __init__(self):
        self._model_dict: dict[str, callable] = {}
        self._table_structure_dict: dict[str, list[TableField]] = {}

    def get(self, data_source: IDataSource, database: str, table: str) -> callable:
        data_source_id = data_source.get_id()
        key = f'{data_source_id}.{database}.{table}'
        model = self._model_dict.get(key, None)
        if model is None:
            conn = data_source.get_connection()
            c = conn.cursor()
            try:
                sql = f'show columns from {database}.{table}'
                logger.debug(f'[{id(conn)}] {sql}')

                c.execute(sql)
                rows = c.fetchall()
                table_fields = []
                for row in rows:
                    table_field = TableField(field_=row['Field'],
                                             type_=row['Type'],
                                             null_=row['Null'],
                                             key_=row['Key'],
                                             default_=row['Default'],
                                             extra=row['Extra'])
                    table_fields.append(table_field)
                self._table_structure_dict[key] = table_fields
                fields = [(table_field.field_, any, field(default=None)) for table_field in table_fields]
                self._model_dict[key] = make_dataclass(key, fields=fields)
                return self._model_dict[key]
            finally:
                c.close()
                conn.commit()
        else:
            return model


class TableField:

    def __init__(self, field_, type_, null_, key_, default_, extra):
        self.field_ = field_
        self.type_ = type_
        self.null_ = null_
        self.key_ = key_
        self.default_ = default_
        self.extra = extra


models = Models()



import time

import pymysql
from loguru import logger
from pymysql.connections import Connection
from pymysql.cursors import DictCursor


class MysqlDataSource:
    def __init__(self, id_, host, port, user, password, database):
        self._id = id_
        self._dialect = 'mysql'
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._default_database = database
        self._conn: Connection | None = None
        self._conn_create_time = time.time()

    def get_id(self) -> str:
        return self._id

    def get_host(self) -> str:
        return self._host

    def get_port(self) -> int:
        return self._port

    def get_user(self) -> str:
        return self._user

    def get_password(self):
        return self._password

    def get_default_database(self) -> str:
        return self._default_database

    def get_connection(self):
        if self._conn is None:
            return self._create_connection()

        # ping every 60 seconds and reconnect if necessary
        if time.time() - self._conn_create_time > 60:
            self._conn.ping(True)
            self._conn_create_time = time.time()
        return self._conn

    def _create_connection(self):
        conn = pymysql.connect(host=self._host,
                               port=self._port,
                               user=self._user,
                               password=self._password.encode('utf-8'),
                               database=self._default_database,
                               cursorclass=DictCursor)
        logger.info(f'create connection [{id(conn)}]')

        self._conn = conn
        self._conn_create_time = time.time()
        return self._conn



from loguru import logger

from ._mysql_datasource import MysqlDataSource
from .protocols import IDataSource


class Dorm:
    def __init__(self):
        self._init = False
        self._config_dict = None
        self._datasource = []

    def is_initialized(self):
        return self._init

    def init(self, config_dict):
        self._config_dict = config_dict

        data_source_config = self.get_config('data_source')
        for ds_id, ds_conf in data_source_config.items():
            if 'dialect' not in ds_conf:
                raise ValueError('dialect is required')

            if 'mysql' == ds_conf['dialect']:
                mysql_ds = MysqlDataSource(id_=ds_id,
                                           host=ds_conf['host'],
                                           port=ds_conf['port'],
                                           user=ds_conf['user'],
                                           password=ds_conf['password'],
                                           database=ds_conf['database'])
                self._datasource.append(mysql_ds)
            else:
                raise ValueError(f'unsupported dialect: {ds_conf["dialect"]}')

        self._init = True
        logger.info('dorm initialized')

    def get_config(self, *keys):
        conf = self._config_dict
        if not conf:
            raise ValueError('uninitialized')
        for key in keys:
            conf = conf.get(key)
            if not conf:
                raise ValueError(f'config not found: {keys}')
        return conf

    def default_datasource(self) -> IDataSource | None:
        for ds in self._datasource:
            if ds.get_id() == 'default':
                return ds
        return None

    def begin(self):
        raise NotImplementedError()

    def commit(self):
        raise NotImplementedError()


dorm = Dorm()



from ._dorm import dorm
from ._initializer import init
from ._insert import insert, insert_bulk, upsert, upsert_bulk
from ._entity import entity
from ._query import query, raw_query
from ._update import update

__author__ = 'melon'
__version__ = '0.3.0'

__all__ = [
    'init',
    'dorm',
    'query',
    'update',
    'insert',
    'insert_bulk',
    'upsert',
    'upsert_bulk',
    'raw_query',
    'entity'
]



import setuptools

setuptools.setup(
    name='pydorm',
    version='0.3.0',
    description="A dynamic and lightweight Python orm framework",
    author="melon",
    packages=setuptools.find_packages(),
)
